package sim.app.socialagents;

import sim.engine.*;
import sim.field.continuous.*;
import sim.util.*;
import ec.util.*;
import java.util.*;
import bioutil.Morphogen;

public class CommSocialSim extends SimState {

	public int fieldWidth, fieldHeight;

	public int numAgents, agentSenseDist;

	public double agentSpeed;

	public Continuous2D agents;

	public Continuous2D obstacles;

	public Continuous2D areasOfInterest;

	public ConfigParams params;

	protected CommDataCollector dataCollector;

	public AgentController agentControl;
	
	public double commParams[];
	
	Agent agentsList[];
	
	public Double2D agentInitPosition;
	public Double2D agentInitVelocity;
	public int agentSize;
	
	public CommSocialSim(long seed, AgentController controller, int runLength,
			int runIntervals, String paramFileName, double[] commParams) {

		super(new MersenneTwisterFast(seed), new Schedule(1));

		this.params = new ConfigParams(paramFileName);

		this.fieldWidth = this.params.getParamInt("FIELD_WIDTH");
		this.fieldHeight = this.params.getParamInt("FIELD_HEIGHT");
		this.numAgents = this.params.getParamInt("NUM_AGENTS");
		this.agentSenseDist = this.params.getParamInt("AGENT_SENSOR_RADIUS");
		this.agentSpeed = this.params.getParamDouble("AGENT_SPEED");

		if (runLength > 0) {
			//this.dataCollector = new CommDataCollector(runLength, runIntervals);
		} else {
			//this.dataCollector = null;
		}

		this.agentControl = controller;
		this.commParams=commParams;
		agentsList= new CommSocialAgent[this.numAgents];
	}

	public CommSocialSim(long seed) {
		this(seed, null, 0, 0, "run.conf", null);
	}

	public void start() {
		
		super.start();
		agentSize = this.params.getParamInt("AGENT_SIZE");
		double deploymentInterval = this.params
				.getParamDouble("DEPLOYMENT_INTERVAL");
		agentInitPosition = this.params
				.getParamDouble2D("AGENT_INITIAL_POSITION");
		agentInitVelocity = this.params
				.getParamDouble2D("AGENT_INITIAL_VELOCITY");
		int[] obs = this.params.getParamIntArray("OBSTACLES");
		int[] aois = this.params.getParamIntArray("AOIS");
		boolean useSocialAgents = this.params.getParamBoolean("SOCIAL_AGENTS");
		int maxScale;

		// add the obstacles to the simulation
		// first find maximum scale so we can set the discretization
		maxScale = 50;
		if (obs != null) {
			for (int i = 0; i < obs.length; i += 4) {
				maxScale = Math.max(maxScale, obs[i + 3]);
			}
		}

		this.obstacles = new Continuous2D(maxScale * 2, this.fieldWidth,
				this.fieldHeight);
		if (obs != null) {
			for (int i = 0; i < obs.length; i += 4) {
				this.obstacles.setObjectLocation(new Obstacle(obs[i],
						obs[i + 1], obs[i + 2], obs[i + 3]), new Double2D(
						obs[i], obs[i + 1]));
			}
		}

		// add the AOIs to the simulation
		maxScale = 50;
		if (aois != null) {
			for (int i = 0; i < aois.length; i += 4) {
				maxScale = Math.max(maxScale, aois[i + 3]);
			}
		}

		this.areasOfInterest = new Continuous2D(maxScale * 2, this.fieldWidth,
				this.fieldHeight);
		if (aois != null) {
			for (int i = 0; i < aois.length; i += 4) {
				AreaOfInterest aoi = new AreaOfInterest(aois[i], aois[i + 1],
						aois[i + 2], aois[i + 3], i / 4);
				this.areasOfInterest.setObjectLocation(aoi, new Double2D(
						aois[i], aois[i + 1]));
			}
		}

		// put the agents in the xsimulation
		Agent newAgent;
		
		this.agents = new Continuous2D(this.agentSenseDist, this.fieldWidth,
				this.fieldHeight);

		for (int i = 0; i < this.numAgents; i++) {
			if (useSocialAgents) {
				//System.out.println("Using social agents");
				newAgent = new CommSocialAgent(agentInitVelocity,
						agentInitPosition, this.agentSenseDist, agentSize,
						this.agentSpeed, this.random, this.agentControl, this);
				
			} else {
				//System.out.println("Using nonsocial agents");
				newAgent = new NonSocialAgent(agentInitVelocity,
						agentInitPosition, this.agentSenseDist, agentSize,
						this.agentSpeed, this.random, this.agentControl);
			}
			agentsList[i]=newAgent;
			this.agents.setObjectLocation(newAgent, agentInitPosition);
			this.schedule.scheduleRepeating((deploymentInterval * i)+10, newAgent);
		}

		if (this.dataCollector != null) {
			this.schedule.scheduleRepeating(0, this.dataCollector);
		}
	}

	public void finish() {
		if (this.dataCollector != null) {
			this.dataCollector.printData();
		}
		for(int i=0; i < this.numAgents; i++){
			((CommSocialAgent)agentsList[i]).killThreads();
		}
	}

	public static void main(String[] args) {
		doLoop(SocialSim.class, args);
		System.exit(0);
	}

	


	/*
	 * returns the number of agents that are currently on top of an AOI
	 */
	public int numAgentsOnAOIs() {

		final Bag allAgents = this.agents.getAllObjects();
		int numOnAOI = 0;

		// iterate through each agent
		for (int i = 0; i < allAgents.numObjs; i++) {

			final Agent agent = (Agent) (allAgents.objs[i]);

			if (agent.getOnAOI())
				numOnAOI++;

		}

		return numOnAOI;

	}

	/*
	 * returns the percentage of AOIs that are currently covered by the Agents
	 */
	public double percentageAOIsCovered() {

		final Bag agents = this.agents.getAllObjects();
		int numAOIs = this.areasOfInterest.getAllObjects().numObjs;
		boolean[] aoiIsCovered = new boolean[numAOIs];
		int numAOIsCovered;

		Arrays.fill(aoiIsCovered, false);

		for (int i = 0; i < agents.numObjs; i++) {

			final Agent agent = (Agent) (agents.objs[i]);

			if (agent.getOnAOI()) {
				aoiIsCovered[agent.aoiCurrentlyObserving()] = true;
			}

		}

		numAOIsCovered = 0;
		for (int i = 0; i < aoiIsCovered.length; i++) {
			if (aoiIsCovered[i])
				numAOIsCovered++;
		}

		return (double) numAOIsCovered / numAOIs;

	}

	/*
	 * returns the average number of unique AOIs seen by the agents
	 */
	public double avgAOIsSeenByAgents() {

		final Bag agents = this.agents.getAllObjects();

		int[] aoisSeen = new int[agents.numObjs];

		for (int i = 0; i < agents.numObjs; i++) {

			final Agent agent = (Agent) (agents.objs[i]);

			aoisSeen[i] = agent.getAOIsSeen();

		}

		return UtilityFunctions.getArrayAvg(aoisSeen);

	}

	/*
	 * returns the average percentage of time that the agents are on AOIs
	 */
	public double avgPercentTimeOverAOIs() {

		final Bag agents = this.agents.getAllObjects();
		double[] avgTimeOverAOI = new double[agents.numObjs];

		for (int i = 0; i < agents.numObjs; i++) {

			final Agent agent = (Agent) (agents.objs[i]);

			if (this.schedule.getSteps() > 0) {
				avgTimeOverAOI[i] = (double) agent.getTimeOverAOI()
						/ this.schedule.getSteps();
			} else {
				avgTimeOverAOI[i] = 0.0;
			}

		}

		return UtilityFunctions.getArrayAvg(avgTimeOverAOI);

	}

	/*
	 * Setters and getters
	 */
	public int getNumAgents() {
		return this.numAgents;
	}

	public void setNumAgents(int val) {
		if (val >= 1)
			this.numAgents = val;
	}

	public double getAgentSpeed() {
		return this.agentSpeed;
	}

	public void setAgentSpeed(double val) {
		if (val >= 0.0)
			this.agentSpeed = val;
	}

	public int getSensorDist() {
		return this.agentSenseDist;
	}

	public void setSensorDist(int val) {
		if (val >= 0)
			this.agentSenseDist = val;
	}

}
