package sim.app.socialagents;

import java.text.*;
import java.util.*;
import java.io.*;

import bioutil.Morphogen;
import sim.engine.*;
import sim.util.Bag;
import ec.util.*;

public class GeneticAlgorithm{

	private int numGenerations;
	private int popSize;
	private double xoverRate, mutRate;

	private ConfigParams params;
	private MersenneTwisterFast rng;

	private String paramFileName;
	private int runNum;
	private long rngSeed;
	private int runLength, recordIntervals;

	// GA data
	private SingleControl[][] population;
	private double[] fitness;

	// statistical GA data
	private boolean recordDetailedData;
	private double[] avgFitness, stdDevFitness, bestFitness;
	private double[][] simRunDataMax;
	private double[][] simRunDataLast;

	// output files
	private PrintWriter runWriter;

	public GeneticAlgorithm(String runConfName){

		this.paramFileName = runConfName;

		this.params = new ConfigParams(runConfName);

		/* if the random number seed is in the parameter file, then use it
		 * otherwise, create a random seed from the system clock
		 **/
		this.rngSeed = this.params.getParamLong("RNG_SEED");
		if(this.rngSeed <= 0){
			this.rngSeed = System.currentTimeMillis();
		}
		this.rng = new MersenneTwisterFast(this.rngSeed);

		this.runNum = this.getCurrentRunNum();
		this.createOutputDirectories();
		this.writeNextRunNum();
		this.initRunWriter();
		this.params.printParamsFile("output/run" + this.runNum + "/params.txt");

		this.recordDetailedData = params.getParamBoolean("RECORD_DETAILED_DATA");

		// get the basic GA parameters from the parameter file
		this.numGenerations = this.params.getParamInt("NUM_GENERATIONS");
		this.popSize        = this.params.getParamInt("POP_SIZE");
		this.xoverRate      = this.params.getParamDouble("XOVER_RATE");
		this.mutRate        = this.params.getParamDouble("MUTATION_RATE");

		this.population = new SingleControl[this.popSize][(int)(Math.pow(2, 0))];
		this.fitness = new double[this.popSize];

		for(int i = 0; i < this.population.length; i++){

			for(int j = 0; j < this.population[i].length; j++){

				this.population[i][j] = new SingleControl(this.rng, this.params.getParamDouble("STD_DEV_GAUSSIAN_MUTATION"));
			}

		}

		this.avgFitness = new double[this.numGenerations];
		this.stdDevFitness = new double[this.numGenerations];
		this.bestFitness = new double[this.numGenerations];
		this.simRunDataMax = new double[this.popSize][];
		this.simRunDataLast = new double[this.popSize][];

		this.runLength = this.params.getParamInt("RUN_LENGTH");
		this.recordIntervals = this.params.getParamInt("GA_RECORD_INTERVAL");
	}

	public static void main(String[] args){

		GeneticAlgorithm ga;
		AgentController controller;
		SocialSim sim;
		long steps;

		if(args.length == 0){
			ga = new GeneticAlgorithm("run.conf");
		}else{
			ga = new GeneticAlgorithm(args[0]);
		}

		if(ga.params.getParamBoolean("SOCIAL_AGENTS")){
			controller = new SocialAgentController(ga.rng);
		}else{
			controller = new NonSocialAgentController(ga.rng);
		}
		
		// main GA loop
		for(int i = 0; i < ga.numGenerations; i++){

			System.out.println("Gen: " + i);
			System.out.print("Individual:");

			for(int j = 0; j < ga.popSize; j++){

				System.out.print(" " + j);

				for(int curCond = 0; curCond < ga.population[j].length; curCond++){
					controller.addControl(ga.population[j][curCond]);
				}

				// create a brand new simulation and run it
				sim = new SocialSim(ga.rngSeed + ((i * ga.numGenerations) + j), controller, ga.runLength, ga.recordIntervals, ga.paramFileName);

				steps = 0;
				sim.start();
				
				while(steps <= ga.runLength){

					if(!sim.schedule.step(sim)){
						break;
					}

					if(steps==30){
						
						
//						 Testing gradients 
						final Bag agents = sim.agents.getAllObjects();
						final SocialAgent agent = (SocialAgent)(agents.objs[0]);
						agent.createGradient(new Morphogen("A",4),sim);
					}
					steps = sim.schedule.getSteps();

					//if(steps % 2500 == 0) System.out.print(" " + steps);

				}
				

				ga.fitness[j] = sim.obtainFitnessForGA();

				ga.runWriter.print("Generation " + i);
				ga.runWriter.print(" individual " + j);
				ga.runWriter.print(" RNG_Seed " + (ga.rngSeed + ((i * ga.popSize) + j)));
				ga.runWriter.print(" Fitness " + ga.fitness[j]);

				for(int curCond = 0; curCond < ga.population[j].length; curCond++){
					ga.runWriter.print(" " + ga.population[j][curCond].getString());
				}

				ga.runWriter.println();

				ga.simRunDataMax[j] = sim.dataCollector.getMaxOfDataValues();
				ga.simRunDataLast[j] = sim.dataCollector.getLastRowOfData();

			}

			System.out.println();

			/* records the full list of statistical data for this generation
			 **/
			if(ga.recordDetailedData){
				ga.recordDetailedGenerationOutput(i);
			}

			ga.avgFitness[i] = UtilityFunctions.getArrayAvg(ga.fitness);
			ga.stdDevFitness[i] = UtilityFunctions.getArrayStdDev(ga.fitness);
			ga.bestFitness[i] = UtilityFunctions.getArrayMaxValue(ga.fitness);

			ga.recombinePop();

		}

		ga.recordRunOutput();

		ga.runWriter.close();

		System.exit(0);

	}

	private void recombinePop(){

		SingleControl[][] tmpPop = new SingleControl[this.popSize][];
		int selectedIndex;

		// select the parents to be in the next generation
		for(int i = 0; i < tmpPop.length; i++){

			selectedIndex = this.selectChild();

			tmpPop[i] = new SingleControl[this.population[selectedIndex].length];

			for(int j = 0; j < tmpPop[i].length; j++){
				tmpPop[i][j] = new SingleControl(this.population[selectedIndex][j]);
			}

		}

		// crossover
		for(int i = 0; i < tmpPop.length; i+=2){

			if(this.rng.nextDouble() <= this.xoverRate){
				crossover(tmpPop[i], tmpPop[i+1]);
			}

		}

		// mutate
		for(int i = 0; i < tmpPop.length; i++){

			for(int j = 0; j < tmpPop[i].length; j++){

				for(int k = 0; k < SingleControl.CONTROL_LENGTH; k++){

					if(this.rng.nextDouble() < this.mutRate){
						tmpPop[i][j].mutate(k, this.rng);
					}

				}

			}

		}

		// put the new individuals back in the population
		for(int i = 0; i < this.population.length; i++){

			this.population[i] = new SingleControl[tmpPop[i].length];

			for(int j = 0; j < this.population[i].length; j++){
				this.population[i][j] = new SingleControl(tmpPop[i][j]);
			}

		}
			
	}

	/* use a simple tournament selection method
	 * returns the index of the selected individual
	 **/
	private int selectChild(){
		
		int index1, index2;

		index1 = this.rng.nextInt(this.popSize);

		// make sure index2 does not equal index1
		index2 = this.rng.nextInt(this.popSize);
		while(index1 == index2) index2 = this.rng.nextInt(this.popSize);

		if(fitness[index1] > fitness[index2]){
			return index1;
		}else{
			return index2;
		}

	}

	/* use a simple roulette wheel selection method
	 * returns the index of the selected individual
	 **/
	private int selectChild2(){
		
		double value = this.rng.nextDouble() * UtilityFunctions.getArraySum(this.fitness);
		int index = 0;

		while(fitness[index] <= value){

			value -= this.fitness[index];
			index++;

		}

		return index;

	}

	/* performs a two point crossover
	 **/
	private void crossover(SingleControl[] parent1, SingleControl[] parent2){

		int xOverPoint1 = this.rng.nextInt(SingleControl.CONTROL_LENGTH * parent1.length);
		int xOverPoint2 = this.rng.nextInt(SingleControl.CONTROL_LENGTH * parent1.length);

		if(xOverPoint1 <= xOverPoint2){
			SingleControl.crossover(parent1, parent2, xOverPoint1, xOverPoint2);
		}else{
			SingleControl.crossover(parent1, parent2, xOverPoint2, xOverPoint1);
		}

	}

	/* writes out the fitness details for the current run
	 **/
	private void recordRunOutput(){

		PrintWriter printWriter;
		DecimalFormat df = new DecimalFormat("0.000000000");

		try{

			printWriter = new PrintWriter(new BufferedWriter(new FileWriter("output/run" + this.runNum + "/fitness.txt")));

		}catch(IOException e){
			System.err.println("(GeneticAlgorithm2.recordDetailedOutput) ERROR: An IOException was thrown while opening output/run" + this.runNum + "/runDetails.txt");
			return;
		}

		printWriter.println();
		printWriter.println("GEN      AVG_FITNESS          STD_DEV_FITNESS      BEST_FITNESS");

		for(int i = 0; i < this.numGenerations; i++){

			printWriter.print(UtilityFunctions.leftPad(i, 3));
			printWriter.print("      ");
			printWriter.print(UtilityFunctions.leftPad(df.format(this.avgFitness[i]), 16));
			printWriter.print("     ");
			printWriter.print(UtilityFunctions.leftPad(df.format(this.stdDevFitness[i]), 16));
			printWriter.print("     ");
			printWriter.print(UtilityFunctions.leftPad(df.format(this.bestFitness[i]), 16));
			printWriter.println();

		}
		
		printWriter.close();

	}

	/* for the current generation, the data from the simulator's DataCollection
	 *   object is written to its own file
	 **/
	private void recordDetailedGenerationOutput(int generation){

		PrintWriter printWriter;
		DecimalFormat df = new DecimalFormat("0.00000");

		try{

			printWriter = new PrintWriter(new BufferedWriter(new FileWriter("output/run" + this.runNum + "/gen" + generation + ".txt")));

		}catch(IOException e){
			System.err.println("(GeneticAlgorithm2.recordDetailedOutput) ERROR: An IOException was thrown while opening output/run" + this.runNum + "/gen" + generation + ".txt");
			return;
		}

		printWriter.println("Max values from each individual's DataCollector:");
		printWriter.println(DataCollector.getHeader());
		for(int i = 0; i < this.popSize; i++){
			printWriter.println(DataCollector.getString(this.simRunDataMax[i]));
		}

		printWriter.println();
		printWriter.println("Last values from each individual's DataCollector:");
		printWriter.println(DataCollector.getHeader());
		for(int i = 0; i < this.popSize; i++){
			printWriter.println(DataCollector.getString(this.simRunDataLast[i]));
		}

		printWriter.close();
		
	}

	/* reads the run.num file and returns the value in that file
	 * if the file does not exist, this function returns 0
	 **/
	private int getCurrentRunNum(){

		try{

			BufferedReader bufReader = new BufferedReader(new FileReader("output/run.num"));
			return Integer.parseInt(bufReader.readLine().trim());

		}catch(IOException e){

			return 0;

		}

	}

	/* write the incremented run number to output/run.num
	 **/
	private void writeNextRunNum(){

		try{

			PrintWriter printWriter = new PrintWriter(new BufferedWriter(new FileWriter("output/run.num")));
			printWriter.println(this.runNum+1);
			printWriter.close();

		}catch(IOException e){
			System.err.println("(GeneticAlgorithm2.writeNextRunNum) ERROR: An IOException was thrown while writing output/run.num");
		}

	}

	/* first creates the output directory (which may already exist)
	 * then creates the directory to hold all of the files for this particular run
	 **/
	private void createOutputDirectories(){

		(new File("output")).mkdir();
		(new File("output/run" + this.runNum)).mkdir();

	}

	/* initializes the runWriter for this run
	 * the runWriter will write out the individuals for each generation, along with thier fitness
	 * the user can have this info written to the screen if they edit the appropriate variable in run.conf
	 **/
	private void initRunWriter(){

		try{

			if(this.params.getParamBoolean("WRITE_GA_RUN_TO_SCREEN")){
				this.runWriter = new PrintWriter(new OutputStreamWriter(System.out), true);
			}else{
				this.runWriter = new PrintWriter(new BufferedWriter(new FileWriter("output/run" + this.runNum + "/fullRun.txt")));
			}

			this.runWriter.println("#DATE_RAN: " + new Date().toString());
			this.runWriter.println("#RNG_SEED: " + this.rngSeed);
			this.runWriter.println();

		}catch(IOException e){
			System.err.println("(GeneticAlgorithm2.initRunWriter) ERROR: An IOException was thrown while opening the runWriter");
		}
	}

}
