/* Author: Adam Campbell
 *
 * This code was modified from the Region.java file in the app/mav directory
 *
 **/

package sim.app.socialagents;

import sim.portrayal.*;
import sim.util.*;
import java.awt.geom.*;
import java.awt.*;

public abstract class Region extends SimplePortrayal2D{

	// the x and y vertice list for the polygon
	protected int[] xPoints;
	protected int[] yPoints;
	private MutableDouble2D center; // the center of this Region
	protected int anchorX, anchorY; // the upper left hand point of the Region
	protected int scaleX, scaleY;

	private Shape shape;
	protected Color color;

	private boolean active, hasChanged;


	/* ax, ay - anchorX and anchorY; respectively
	 * sx, sy - scaleX and scaleY; respectively
	 **/
	public Region(int ax, int ay, int sx, int sy){

		this.anchorX = ax;
		this.anchorY = ay;
		this.scaleX = sx;
		this.scaleY = sy;

		this.center = new MutableDouble2D(ax + sx/2.0, ay + sy/2.0);
		this.resetPolygon();
		
		this.shape = new Polygon(this.xPoints, this.yPoints, this.xPoints.length);

		this.active = true;
		this.hasChanged = false;
	}

	private void resetPolygon(){

		this.xPoints = new int[]{0, 1, 1, 0};
		this.yPoints = new int[]{0, 0, 1, 1};

		for(int i = 0; i < this.xPoints.length; i++){
			this.xPoints[i] = (this.xPoints[i] * this.scaleX) + this.anchorX;
			this.yPoints[i] = (this.yPoints[i] * this.scaleY) + this.anchorY;
		}

		this.shape = new Polygon(this.xPoints, this.yPoints, this.xPoints.length);

		this.hasChanged = true;
	}

	/* The code below is taken from the demo.  The authors of the original code know what they're
	 *   doing.  Therefore there's no reason for me to assume there's a better way of writing this
	 *   function.
	 *
	 * rule 1: don't fool around with graphics' own transforms because they effect its clip, ARGH.
	 * so we have to create our own transformed shape.  To be more efficient, we only transform
	 * it if it's moved around.
	 **/
	Shape oldShape;
	Rectangle2D.Double oldDraw = null;
	public void draw(Object object, Graphics2D graphics, DrawInfo2D info){

		/* check to see if the object has been moved, or if the scale has changed
		 * if either of the above have occured, then reset the shape info
		 **/
		if (this.oldDraw == null || this.oldDraw.x != info.draw.x || this.oldDraw.y != info.draw.y || 
		    this.oldDraw.width != info.draw.width || this.oldDraw.height != info.draw.height || this.hasChanged){
			this.oldDraw = info.draw;
			AffineTransform transform = new AffineTransform();
			transform.translate(0.0,0.0);//this.oldDraw.x, this.oldDraw.y);
			transform.scale(this.oldDraw.width, this.oldDraw.height);
			this.oldShape = transform.createTransformedShape(shape);
			this.hasChanged = false;
		}
		
		// draw the shape
		graphics.setColor(this.color);

		// if it's not active, just draw its outline
		if(this.active){
			graphics.fill(this.oldShape);
		}else{
			graphics.draw(this.oldShape);
		}

	}

	/** If drawing area intersects selected area, add to the bag */
	public void hitObjects(DrawInfo2D range, Bag putInHere){

		AffineTransform transform = new AffineTransform();
		transform.translate(range.draw.x, range.draw.y);
		transform.scale(range.draw.width, range.draw.height);
		Shape s = transform.createTransformedShape(shape);
				
		if (s.intersects(range.clip.x, range.clip.y, range.clip.width, range.clip.height))
			putInHere.add(this);

	}

	public boolean hitObject(Object object, DrawInfo2D range){

		int x, y;

		x = (int)(range.clip.x/range.draw.width);
		y = (int)(range.clip.y/range.draw.height);

		return this.pointInsideRegion(x, y);

	}

	/* returns a boolean indicating whether or not the point (x, y)
	 *   is located inside the polygon
	 **/
	public boolean pointInsideRegion(double x, double y){

		return ((Polygon)this.shape).contains(x, y);

	}

	/* returns a Double2D containing the closest point on the region to the given point
	 * if the distance to the closest point is greater than maxDist, then new Double2D(Double.MAX_VALUE, Double.MAX_VALUE) is returned
	 **/
	public Double2D getClosestPoint(double x, double y, double maxDist){

		Bag visiblePoints = this.getVisiblePoints(x, y, maxDist);
		double minDist = Double.MAX_VALUE;
		Double2D closestPoint, currentPoint;
		double dist;

		closestPoint = new Double2D(Double.MAX_VALUE, Double.MAX_VALUE);

		for(int i = 0; i < visiblePoints.numObjs; i++){

			currentPoint = (Double2D)visiblePoints.objs[i];

			dist = currentPoint.distance(x, y);

			if(dist < minDist){
				closestPoint = currentPoint;
				minDist = dist;
			}

		}

		return closestPoint;

	}

	/* returns a bag containing all of the closest points on the perimeter of the region within maxDist of (x,y)
	 **/
	public Bag getVisiblePoints(double x, double y, double maxDist){

		int numPoints = xPoints.length;
		Bag visiblePoints = new Bag();
		double distance;

		for(int i = 0; i < numPoints; i++){
			distance = Region.segmentToPointDistance(x,y,xPoints[i], yPoints[i], xPoints[(i+1)%numPoints], yPoints[(i+1)%numPoints]); 
			
			if(distance <= maxDist){
				visiblePoints.add(Region.closestPointOnSegment(x,y,xPoints[i], yPoints[i], xPoints[(i+1)%numPoints], yPoints[(i+1)%numPoints]));
			}
		}
		return visiblePoints;
	}
	
	
	public static double distance(double x1, double y1, double x2, double y2) 
	{
		return Math.sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2)); 
	}

	/* Given the endpoints of a line segment (segmentP0 and segmentP1)
	 *   this function returns the shortest distance from point to the segment
	 *
	 * This algorithm found on http://softsurfer.com/Archive/algorithm_0102/algorithm_0102.htm
	 **/
	private static double segmentToPointDistance(double pX, double pY, double x1, double y1, double x2, double y2)
	{
		double Vx = x2 - x1; 
		double Vy = y2 - y1; 
		double Wx = pX - x1; 
		double Wy = pY - y1; 
		
		double c1 = Vx * Wx + Vy * Wy; 
		double c2 = Vx * Vx + Vy * Vy; 
		
		if(c1 <= 0.0){
			return distance(x1,y1,pX,pY);
		}

		if(c2 <= c1){
			return distance(x2,y2,pX,pY);
		}

		double b = c1/c2;
		return distance(x1+Vx*b, y1+Vy*b, pX,pY);

	}

	/* Given the endpoints of a line segment (segmentP0 and segmentP1)
	 *   this function returns the closest point on the segment from the point
	 **/
	//private static Double2D closestPointOnSegment(MutableDouble2D point, MutableDouble2D segmentP0, MutableDouble2D segmentP1){
	private static Double2D closestPointOnSegment(double pX, double pY, double x1, double y1, double x2, double y2)
		{
		double Vx = x2 - x1; 
		double Vy = y2 - y1; 
		double Wx = pX - x1; 
		double Wy = pY - y1; 
		
		double c1 = Vx * Wx + Vy * Wy; 
		double c2 = Vx * Vx + Vy * Vy; 
		
		if(c1 <= 0.0){
			return new Double2D(x1,y1);
		}

		if(c2 <= c1){
			return new Double2D(x2,y2);
		}

		double b = c1/c2;
		return new Double2D(x1+Vx*b, y1+Vy*b); 
	}

	/* setters and getters
	 **/
	public boolean isActive() { return this.active; }
	public void setActive(boolean b) { this.active = b; }

	public int getAnchorX() { return this.anchorX; }
	public void setAnchorX(int val) { this.anchorX = val; this.resetPolygon(); }
	public int getAnchorY() { return this.anchorY; }
	public void setAnchorY(int val) { this.anchorY = val; this.resetPolygon(); }

	public int getScaleX() { return this.scaleX; }
	public void setScaleX(int val) { this.scaleX = val; this.resetPolygon(); }
	public int getScaleY() { return this.scaleY; }
	public void setScaleY(int val) { this.scaleY = val; this.resetPolygon(); }

	public MutableDouble2D getCenter(){ return this.center; }

	// because we're using Areas, and for some bizarre reason Area isn't serializable,
	// if WE want to be serializable or externalizable we need to handle our own read
	// and write methods.
	
	/*private void writeObject(java.io.ObjectOutputStream p)
		throws IOException
		{
		p.writeDouble(originx);
		p.writeDouble(originy);
		p.writeInt(shapeNum);
		p.writeInt(surface);
		}
		
	private void readObject(java.io.ObjectInputStream p)
		throws IOException, ClassNotFoundException
		{
		originx = p.readDouble();
		originy = p.readDouble();
		shapeNum = p.readInt();
		System.out.println(shapeNum);
		surface = p.readInt();
		// reload shapes and areas, which aren't serializable (ugh)
		shape = (Shape)(shapes[shapeNum]);
		area = new Area(shape);
		}	*/

}
