package sim.app.socialagents;

import ec.util.*;
import java.text.*;

public class SingleControl{

	private final static double DEFAULT_STD_DEV_GAUSSIAN = 0.10;

	public static int CONTROL_LENGTH;

	private boolean[] conditions;
	private double[] parameters;

	private int maxIntValue;
	private double stdDevGaussian;

	public SingleControl(){

		SingleControl.CONTROL_LENGTH = AgentController.NUM_CONDITIONS + AgentController.NUM_PARAMS;

		this.conditions = new boolean[AgentController.NUM_CONDITIONS];
		this.parameters = new double[AgentController.NUM_PARAMS];

		this.stdDevGaussian = SingleControl.DEFAULT_STD_DEV_GAUSSIAN;

	}

	public SingleControl(SingleControl that){

		SingleControl.CONTROL_LENGTH = AgentController.NUM_CONDITIONS + AgentController.NUM_PARAMS;

		this.conditions = new boolean[that.conditions.length];
		this.parameters = new double[that.parameters.length];

		for(int i = 0; i < this.conditions.length; i++){
			this.conditions[i] = that.conditions[i];
		}

		for(int i = 0; i < this.parameters.length; i++){
			this.parameters[i] = that.parameters[i];
		}

		this.stdDevGaussian = that.stdDevGaussian;

	}

	/* creates a randomly initialized SingleControl
	 **/
	public SingleControl(MersenneTwisterFast rng, double stdDevGaus){

		this();

		for(int i = 0; i < this.conditions.length; i++){
			this.conditions[i] = rng.nextBoolean();
		}

		for(int i = 0; i < this.parameters.length; i++){
			this.parameters[i] = rng.nextDouble();
		}

		this.stdDevGaussian = stdDevGaus;
	}

	/* one point crossover
	 **/
	public static void crossover(SingleControl[] control1, SingleControl[] control2, int exchangePoint){

		boolean bTmp;
		double dTmp;
		int controlIndex, paramIndex;

		for(int i = exchangePoint; i < SingleControl.CONTROL_LENGTH * control1.length; i++){

			controlIndex = i / SingleControl.CONTROL_LENGTH;
			paramIndex = i % SingleControl.CONTROL_LENGTH;

			if(paramIndex < AgentController.NUM_CONDITIONS){

				// swap boolean value in conditions array
				bTmp = control1[controlIndex].conditions[paramIndex];
				control1[controlIndex].conditions[paramIndex] = control2[controlIndex].conditions[paramIndex];
				control2[controlIndex].conditions[paramIndex] = bTmp;

			}else{

				paramIndex -= AgentController.NUM_CONDITIONS;

				// swap double value in parameters array
				dTmp = control1[controlIndex].parameters[paramIndex];
				control1[controlIndex].parameters[paramIndex] = control2[controlIndex].parameters[paramIndex];
				control2[controlIndex].parameters[paramIndex] = dTmp;

			}
		}

	}

	/* two point crossover
	 **/
	public static void crossover(SingleControl[] control1, SingleControl[] control2, int exchangePoint1, int exchangePoint2){

		boolean bTmp;
		double dTmp;
		int controlIndex, paramIndex;

		for(int i = exchangePoint1; i <= exchangePoint2; i++){

			controlIndex = i / SingleControl.CONTROL_LENGTH;
			paramIndex = i % SingleControl.CONTROL_LENGTH;

			if(paramIndex < AgentController.NUM_CONDITIONS){

				// swap boolean value in conditions array
				bTmp = control1[controlIndex].conditions[paramIndex];
				control1[controlIndex].conditions[paramIndex] = control2[controlIndex].conditions[paramIndex];
				control2[controlIndex].conditions[paramIndex] = bTmp;

			}else{

				paramIndex -= AgentController.NUM_CONDITIONS;

				// swap double value in parameters array
				dTmp = control1[controlIndex].parameters[paramIndex];
				control1[controlIndex].parameters[paramIndex] = control2[controlIndex].parameters[paramIndex];
				control2[controlIndex].parameters[paramIndex] = dTmp;

			}
		}

	}

	/* location - the index of the value that should be mutated
	 **/
	public void mutate(int location, MersenneTwisterFast rng){

		if(location < AgentController.NUM_CONDITIONS){

			// mutate a value in the conditions array
			this.conditions[location] = !this.conditions[location];

		}else if(location < AgentController.NUM_CONDITIONS + AgentController.NUM_NG_PARAMS){

			location -= AgentController.NUM_CONDITIONS;

			/* mutate a value in the parameters array
			 * a gaussian randomly generated number with mean 0 and standard deviation this.stdDevGaussian
			 * we also make sure that the resulting parameter is in the range of [0, 1]
			 **/
			double value = rng.nextGaussian() * this.stdDevGaussian;

			this.parameters[location] += value;

			if(this.parameters[location] < 0.0) this.parameters[location] = 0.0;
			if(this.parameters[location] > 1.0) this.parameters[location] = 1.0;

		}

	}

	public void setConditionValue(int index, boolean val){

		this.conditions[index] = val;

	}

	public void setParameter(int index, double val){

		this.parameters[index] = val;

	}

	/* returns a string representation of this SingleControl
	 **/
	public String getString(){

		DecimalFormat df = new DecimalFormat("0.00000");
		String ret = "";

		for(int i = 0; i < this.conditions.length; i++){
			ret += (this.conditions[i]?"T":"F");
		}

		ret += "-->";

		for(int i = 0; i < this.parameters.length; i++){
			ret += " " + df.format(this.parameters[i]);
		}

		return ret;

	}


	public boolean[] getConditions(){ return this.conditions; }
	public double[] getParameters(){ return this.parameters; }

}
